/*
  E-Asteroids.c
  Asteroids Epplet 0.1
  Pete Eberlein <eberlein@engr.orst.edu>
  Jan 4, 2000

  Tested with Enlightenment 0.16.3

  Keys:
	Up - Accelerate
        Left, Right - Turn
        Down - Shields
        Space - Shoot

*/

#include "epplet.h"

#define WIDTH 192
#define HEIGHT 192

#define SCALE 16

#define ACTORS 50
#define SHIP 1
#define SHOT 2
#define THRUST 3
#define SCRAP 4
#define SAFE 5
#define SHIELDS 6
#define WRECKED 7
#define ROCK1 11
#define ROCK2 12
#define ROCK3 13

double rnd()
{
    static int rand_max = 0;
    if(!rand_max) {
       	int i=100;
	srand(time(NULL));
	while(i--) rand_max |= rand();
    }
    return (double)rand() / rand_max;
}


static int type[ACTORS],x[ACTORS],y[ACTORS],vx[ACTORS],vy[ACTORS],
    *lines[ACTORS],
    counter[ACTORS],
    dx,dy,
    tx1,ty1,tx2,ty2,tx3,ty3;
static double angle[ACTORS];

static int cant_shoot_timer;
static int level = 0;
static int rocks_left;

static int ship_lines[10];
static int shot_lines[] = {
    -1,0,
    0,1,
    1,0,
    0,-1,
    -1,0,
    0,0};
static int thrust_lines[] = {1,1,1,1,0,0};
static int rock_lines[] = {10,10,9,9,12,12,10,10,9,9,11,11,8,8,10,10,
8,8,11,11,8,8,9,9,12,12,10,10,11,11,10,10};
static int rock_rads[] = {10,7,4};


static void rotate(int actor, double delta)
{
    angle[actor] += delta;
    switch(type[actor]) {

    case SHIP:
    dx = (1 << SCALE) * cos(angle[actor]);
    dy = (1 << SCALE) * sin(angle[actor]);
    
    ship_lines[0] = ship_lines[6] = 5 * cos(angle[actor]);
    ship_lines[1] = ship_lines[7] = 5 * sin(angle[actor]);

    ship_lines[2] = 5 * cos(angle[actor]+2.6);
    ship_lines[3] = 5 * sin(angle[actor]+2.6);

    ship_lines[4] = 5 * cos(angle[actor]-2.6);
    ship_lines[5] = 5 * sin(angle[actor]-2.6);
    break;
    
    case SCRAP:
    lines[actor][0] = -4 * cos(angle[actor]);
    lines[actor][1] = -4 * sin(angle[actor]);

    lines[actor][2] = 4 * cos(angle[actor]);
    lines[actor][3] = 4 * sin(angle[actor]);
    break;
    	
    
    case ROCK1: {
    	int i;
	for (i=0; i < 32; i += 2) {
	    int rad = rock_lines[i] * rock_rads[0] / 10;
	    lines[actor][i] = rad * cos(angle[actor]+i*3.14/16.0);
	    lines[actor][i+1] = rad * sin(angle[actor]+i*3.14/16.0);
	}
	lines[actor][32] = lines[actor][0];
	lines[actor][33] = lines[actor][1];
	lines[actor][34] = 0;
	lines[actor][35] = 0;
	break;
    	}
    case ROCK2: {
    	int i;
	for (i=0; i < 16; i += 2) {
	    int rad = rock_lines[i*2] * rock_rads[1] / 10;
	    lines[actor][i] = rad * cos(angle[actor]+i*3.14/8.0);
	    lines[actor][i+1] = rad * sin(angle[actor]+i*3.14/8.0);
	}
	lines[actor][16] = lines[actor][0];
	lines[actor][17] = lines[actor][1];
	lines[actor][18] = 0;
	lines[actor][19] = 0;
	break;
    	}
    case ROCK3: {
    	int i;
	for (i=0; i < 8; i += 2) {
	    int rad = rock_lines[i*4] * rock_rads[2] / 10;
	    lines[actor][i] = rad * cos(angle[actor]+i*3.14/4.0);
	    lines[actor][i+1] = rad * sin(angle[actor]+i*3.14/4.0);
	}
	lines[actor][8] = lines[actor][0];
	lines[actor][9] = lines[actor][1];
	lines[actor][10] = 0;
	lines[actor][11] = 0;
	break;
    	}
    }
}





static int create(int kind)
{
    int i = 0;
    while ((i < ACTORS-1) && type[i]) i++;
    type[i] = kind;
    return i;
}

static int create_rock(int kind)
{
    int i = create(kind);
    lines[i] = (int*)malloc(sizeof(int) * 
    (kind == ROCK1 ? 36 : (kind == ROCK2 ? 20 : 12)));

    counter[i] = 10*(rnd()-0.5);
    angle[i] = 3.14*2*rnd();
    vx[i] = ((kind-ROCK1+1)<<SCALE) *2* (rnd()-0.5);
    vy[i] = ((kind-ROCK1+1)<<SCALE) *2* (rnd()-0.5);
    rotate(i,0);
        
    return i;    
}


static void
cb_close(void *data)
{
   Epplet_remember();
   Esync();
   exit(0);
}

int Left=0, Right=0, Up=0, Down=0, space=0;

static void
cb_keypress(void *data, Window win, char *key)
{
    if (!strcmp(key, "Left")) Left = 1;
    else if (!strcmp(key, "Right")) Right = 1;
    else if (!strcmp(key, "Up")) Up = 1;
    else if (!strcmp(key, "Down")) Down = 1;
    else if (!strcmp(key, "space")) space = 1;
    else if (!strcmp(key, "Escape")) cb_close(NULL);
/*    else printf("%s key pressed (ignored)\n", key);*/
}

static void
cb_keyrelease(void *data, Window win, char *key)
{
    if (!strcmp(key, "Left")) Left = 0;
    else if (!strcmp(key, "Right")) Right = 0;
    else if (!strcmp(key, "Up")) Up = 0;
    else if (!strcmp(key, "Down")) Down = 0;
    else if (!strcmp(key, "space")) space = 0;
/*    else printf("%s key released (ignored)\n", key);*/
}


static Epplet_gadget da;


static void
cb_timer(void *data)
{
    Window win;
    int i, l;
    
    win = Epplet_get_drawingarea_window(da);
    Epplet_draw_box(win,0,0,WIDTH,HEIGHT,0,0,0);
    
    if (type[0] == SHIP) {
    }

    for (i=0; i < ACTORS; i++) {
    	if (!type[i]) continue;
    	x[i] += vx[i];
    	y[i] += vy[i];
    
    	if (x[i] < 0) x[i] += WIDTH<<SCALE; 
	else if (x[i] >= WIDTH<<SCALE) x[i] -= WIDTH<<SCALE;
    	if (y[i] < 0) y[i] += HEIGHT<<SCALE; 
	else if (y[i] >= HEIGHT<<SCALE) y[i] -= HEIGHT<<SCALE;

    	if(lines[i])
	for(l=0; lines[i][l+2] || lines[i][l+3]; l+=2) {
	    Epplet_draw_line(win,
	    	(x[i]>>SCALE) + lines[i][l],
	    	(y[i]>>SCALE) + lines[i][l+1],
		(x[i]>>SCALE) + lines[i][l+2],
	    	(y[i]>>SCALE) + lines[i][l+3],
		0xFF,0xFF,0xFF);
	}
	
	switch(type[i]) {
	    case SHIP:
        	if (Left) { rotate(i,-0.2); }
    		if (Right) { rotate(i,0.2); }
        	if (Up) { 
		    int j;
        	    double d;
        	    vx[i] += dx/4;
    		    vy[i] += dy/4;
    		    d = (double)vx[i]*vx[i] + (double)vy[i]*vy[i];
		    if (d > 9.0e10) {
			/*printf("too fast! %g\n", d);*/
			d = 3.0e5 / sqrt(d);
			vx[i] *= d;
			vy[i] *= d;
		    }

		    j = create(THRUST);
		    x[j] = x[i] - 5*dx;
		    y[j] = y[i] - 5*dy;
		    vx[j] = vx[i] - 2*dx;
		    vy[j] = vy[i] - 2*dy;
		    counter[j] = 5;
		    lines[j] = (int*)shot_lines;

		}
		if (cant_shoot_timer) cant_shoot_timer--;
		else if (space) {
    		    /*printf("shooting!\n");*/
    		    int j = create(SHOT);
		    x[j] = x[i];
		    y[j] = y[i];
		    vx[j] = vx[i] + 5 * dx;
		    vy[j] = vy[i] + 5 * dy;
		    counter[j] = 25;
		    lines[j] = (int*)shot_lines;

    		    cant_shoot_timer = 3;
		}
		
		if (Down) {
		    type[i] = SHIELDS;
		    lines[i] = NULL;
		    counter[i] = 0;
		}

		break;
		
	    case SAFE: {
	    	int j, safe=1;
	    	for(j=0; j < ACTORS; j++) {
		    if (type[j] >= ROCK1 && type[j] <= ROCK3) {
		    	int dx = (x[i]-x[j]) >> SCALE;
			int dy = (y[i]-y[j]) >> SCALE;
			int rad = rock_rads[type[j]-ROCK1] + 10;
			if(dx*dx+dy*dy <= rad * rad) {
			    safe = 0;
			    break;
			}
		    }
		}
		if (counter[i]) {
		    counter[i]--;
		} else if (safe) {
		    type[i] = SHIP;
		    counter[i] = 0;
		    lines[i] = (int*)ship_lines;
	            rotate(i, 0.0);

		}
		break;
	    }
	    
	    case SHIELDS:
	    	if (!Down) {
		    type[i]=SHIP;
		    lines[i] = (int*)ship_lines;
		    rotate(i,0);
		} else if (counter[i]++ >= 40) {
		    type[i] = WRECKED;
		}
	    	break;

	    case WRECKED:
	    	{
		    int k;
		    for(k=0; k<3; k++) {
		    	int n=create(SCRAP);
			x[n]=x[i];
			y[n]=y[i];
			vx[n]=vx[i]+(3<<SCALE)*(rnd()-0.5);
			vy[n]=vy[i]+(3<<SCALE)*(rnd()-0.5);
			counter[n] = 100;
			angle[n] = 3.14*2*rnd();
			lines[n] = (int*)malloc(6*sizeof(int));
			rotate(n,0);
			lines[n][4] = 0;
			lines[n][5] = 0;
		    }
		    type[i] = SAFE;
    	    	    x[i] = WIDTH / 2 << SCALE;
    	    	    y[i] = HEIGHT / 2 << SCALE;
    	    	    vx[i] = 0;
    	    	    vy[i] = 0;
    	    	    lines[i] = NULL;
		    counter[i] = 120;
    		}
		break;
	    
	    
	    case ROCK1: 
	    case ROCK2: 
	    case ROCK3: 
	    	{
	    	int j, rad = rock_rads[type[i]-ROCK1];
    	    	
	    	for(j=0; j < ACTORS; j++) {
		    if (type[j] == SHOT || type[j] == SHIP) {
		    	int dx = (x[i]-x[j]) >> SCALE;
			int dy = (y[i]-y[j]) >> SCALE;
			int drad = rad + 5*(type[j] == SHIP);
			if(dx*dx+dy*dy <= drad*drad) {
			
			    if(type[j] == SHIP) { 
			    	type[j] = WRECKED; /* wreck the ship */
				vx[j] = (vx[i] + vx[j])/2;
				vy[j] = (vy[i] + vy[j])/2;
			    } else {
			    	type[j] = 0; /* bullets just disappear */
			    }

			    free(lines[i]);

			    if (type[i] != ROCK3) { /* make two smaller rocks */
			    	int a = create_rock(type[i]+1);
			    	int b = create_rock(type[i]+1);
			    	x[a] = x[i];
				y[a] = y[i];
			    	x[b] = x[i];
				y[b] = y[i];
				rocks_left+=2;
			    }

			    type[i] = 0;
			    
			    if (!--rocks_left) {
			    	level++;
				while (rocks_left < level+4) {
        			    i = create_rock(ROCK1);
        			    x[i] = (WIDTH << SCALE) * rnd();
        			    y[i] = (HEIGHT << SCALE) * rnd();
				    rocks_left++;
				}
			    }
				
			    
			    break;
			}
		    }
		}
	    	
	    	rotate(i, counter[i]*0.01);
		break;
	    	}
		
	    case SCRAP: 
	    	rotate(i, 0.1);
	    	if (!counter[i]--) { free(lines[i]); type[i] = 0; }
		break;
		
	    case SHOT:
	    case THRUST:
	    	if (!counter[i]--) type[i] = 0;
		break;
	
	}
	
    }    

    
    Esync();
    Epplet_timer(cb_timer, NULL, 0.05, "TIMER");   
}


static void init()
{
    int i;
    for(i=0; i < ACTORS; i++) {
    	type[i] = 0;
    }
    
    i = create(SAFE);
    x[i] = WIDTH / 2 << SCALE;
    y[i] = HEIGHT / 2 << SCALE;
    vx[i] = 0;
    vy[i] = 0;
    lines[i] = NULL;
    ship_lines[8] = 0;
    ship_lines[9] = 0;
    angle[i] = 0.0;
    counter[i] = 10;
    
    cant_shoot_timer = 0;

    rocks_left = 0;
    while (rocks_left < level+4) {
        i = create_rock(ROCK1);
        x[i] = (WIDTH << SCALE) * rnd();
        y[i] = (HEIGHT << SCALE) * rnd();
	rocks_left++;
    }

}


int
main(int argc, char **argv)
{
   init();

         
   Epplet_Init("E-Asteroids", "0.1", "Enlightenment Asteroids Epplet",
	       WIDTH/16, HEIGHT/16, argc, argv, 0);
   Epplet_timer(cb_timer, NULL, 0.05, "TIMER");

   Epplet_gadget_show(da = Epplet_create_drawingarea(0,0,WIDTH,HEIGHT));
   Epplet_gadget_show(Epplet_create_button(NULL, NULL, 
					   2, 2, 0, 0, "CLOSE", 0, NULL, 
					   cb_close, NULL));

   Epplet_register_key_press_handler(cb_keypress, NULL);
   Epplet_register_key_release_handler(cb_keyrelease, NULL);

   Epplet_show();
   Epplet_Loop();
   return 0;
}
